/*-------------------------------------------------------------
 *  Copyright (c) 2011-2014, MSS ComBlock. All rights reserved.
 *
 * 	Filename: RF.c
 *  Last modified: 4/12/14
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 4
 *  Inheritance: N/A
 *  Description: RF components driver
 *
 *  Rev 1.1 6/1/11 AZ
 *  Moved RF gain control 2 from PWM 1.2 to PWM 1.4 (reason: hardware version change from prototype to production)
 *
 *  Rev 2 4/25/12 AZ
 *  Corrected bug in program_tuner_pll
 *
 *  Rev 3 3/9/13 AZ
 *  Read mixer power detector, even when AGC loop is open
 *  reduced AGC threshold to prevent saturation (wideband noise + signal)
 *  Note: make sure shield is soldered on all 4 edges to prevent oscillations.
 *
 *  Rev 4 4/15/14 AZ
 *  Rewrote AD7298_rw to shave a few us (3 us).
 *  Added ISRs triggered by falling and rising edges of RX_TXN to speed up the switching
 -------------------------------------------------------------*/
#include "main.h"
#include "rf.h"
#include "utils.h"

/*-------- change this section according to the schematics ----------------------------------------*/


/*-------- end of schematics-dependent code ----------------------------------------*/

inline void RX_TXN_R_ISR(){
	// Interrupt service routine when the RX_TX# signal switches to 1
	// called at rising edge.
	// half-duplex operation, receive enabled, tx disabled
	PIN_RX_TXN2_H();	// receive path selected
	PIN_TX_RXN2_L();
	// minimize tx gain in half-duplex operation
	PWM11(0);
}
inline void RX_TXN_F_ISR(){
	// Interrupt service routine when the external RX_TX# signal switches to 0
	// called at falling edge.
	// half-duplex operation, tx enabled, rx disabled
	PIN_RX_TXN2_L();	// transmit path selected
	PIN_TX_RXN2_H();
	// restore tx gain in half-duplex operation
	// change PWM11
	PWM11(tx_gain >> 10);
}

// poll signals received over the flat flexible cable
// Only poll the signals not changing often, TX_EN and RFOUT_ENB,
// The fast changing RX_TX# is covered separately by a fast-reaction interrupt service routine)
// Enable only when regv[0] bit 7 is set
void FFC_signals_poll(){
	if(regv[0] & 0x0080){
		// control lines from FFC enabled
		if(FIO0PIN & 0x00010000){	// TX_EN control
			PIN_TX_EN2_L();		// active low
		} else {
			PIN_TX_EN2_H();
		}
		if(FIO0PIN & 0x00020000){	// RFOUT_ENB
			PIN_RFOUT_ENB2_H();
		} else {
			PIN_RFOUT_ENB2_L();		// active low
		}
	}
}

void DAC_init(void) {
	// initialize internal 10-bit DAC

	// 1. to enable the DAC, output must be selected to appear on the related pin p0.26
	PINSEL1 = (PINSEL1 & 0xFFCFFFFF) | 0x00200000;

	// 2. peripheral clock selection for DAC.
	// select the processor clock CCLK/4 (default)
	PCLKSEL1 = PCLKSEL1 & 0xFF3FFFFF;

	// 3. enable double buffering, no interrupts, 1us refresh
	DACCTRL = 0x00000006;
	DACCNTVAL = 30; // 1us @120MHz/4
}

void DAC(long value) {
	// output is value * Vrefp / 1024
	// 1us settling time -> 1 MHz maximum update rate
	DACR = (value << 6) & 0x0000FFC0;
}

void PWM_init(void) {
	// initialized P1[18] = PWM 1.1 as a fake 10-bit DAC (externally filtered)
	// PWM1IR interrupt register ignored (no interrupts)
	PCONP |= (1 << 6); // PWM1 power ON
	// PCLKSEL0: by default peripheral clock is CCLK/4 = 30 MHz
	PWM1TCR = 0x00000002; // Reset counter
	PWM1PR = 0;
	PWM1PC = 0;
	PINSEL3 = (PINSEL3 & (~0x00000020)) | 0x00000020;
	// p1[18] = PWM
	PINMODE3 = (PINMODE3 & (~0x00000020)) | 0x00000020;
	// P1[18] PWM has no PU/PD
	PWM1MR0 = 1024; // controls the PWM cycle rate. Here 1024 peripheral clocks
	PWM1MR1 = 0; // initial PWM value
	PWM1MR2 = 0; // initial PWM value
	// PWM Match Control Register
	// Match 0 is configured to reset the timer/counter when a match event occurs.
	PWM1MCR = 0x00000002; // reset counter when match MR0
	// PWM Control Register
	PWM1PCR = 0x00000200; // all single edge, PWM1 output enabled
	// PWM Timer Control Register
	PWM1TCR = 0x00000009; // clear counter,  enable PWM mode, enable counting
	// output is value * 3.3V / 1024 after filtering
	PWM1MR0 = 1024; // controls the PWM cycle rate. Here 1024 peripheral clocks
	// PWM Latch Enable Register
	PWM1LER = 0x00000007; // Enable PWM Match 0,1,2 Latch
}

void PWM11(long value) {
	// output is value * 3.3V / 1024 after filtering
	PWM1MR1 = value & 0x000003FF; // controls the PWM edge position
	// PWM Latch Enable Register
	PWM1LER = 0x00000007; // Enable PWM Match 0,1,2 Latch
}


void ALC_poll(void){
	// transmit gain control
	int tx_gain_prev;

	// freeze ALC loop if half-duplex and RX_TXN (P0[18]) in receive mode
	if ((configSw.mid[5] == 'A') && (FIO0PIN & 0x00040000)){
		return;
	}

	if(regv[1] & 0x0001){
		// ALC enabled
		tx_gain_prev = tx_gain;	// value before any change
		if(tx_power_det > ALC_target){
			// too high. reduce gain
			if(tx_gain > 0){
				tx_gain--;
				if((tx_gain_prev & 0x000FFC00) == (tx_gain & 0x000FFC00)){
					// no change to the 10-bit PWM value
					return;
				} else {
					// change PWM11
					PWM11(tx_gain >> 10);
				}


			}
		} else {
			// too low. increase gain
			if(tx_gain != 0x0FFFFF){
				tx_gain++;
				if((tx_gain_prev & 0x000FFC00) == (tx_gain & 0x000FFC00)){
					// no change to the 10-bit PWM value
					return;
				} else {
					// change PWM11
					PWM11(tx_gain >> 10);
				}
			}
		}
	}
}

//void setRFGain(void) {
//	AD5621_1(RF_gain);
//}
//
//void setIFGain(void) {
//	AD5621_2(IF_gain);
//
//}

// call once after tx is enabled
//inline void setTxGain(void){
//	int c;
//	if((regv[39] & 0x30) == 0){
//		// tx gain set by control register. fixed
//		// transmitter gain (10-bit)
//		c = (int)(regv[38]) & 0x00000003;
//		c <<= 8;
//		c |= (int)(regv[37]) & 0x000000FF;
//		DAC(c);
//	} else if ((regv[39] & 0x30) == 0x10){
//		// tx gain set by external analog signal TX_GAIN_CTRL1
//		c = aux_adc[5];
////		c = (c >> 2) & 0x03FF;	// reduce to 10-bit
//		c >>= 2;	// reduce to 10-bit
//		DAC(c);
//	}
//}
//
//// call periodically to track TX_GAIN_CTRL1
//void tx_gain_poll(void){
//	int c;
//	if ((regv[39] & 0x30) == 0x10){
//		// tx gain set by external analog signal TX_GAIN_CTRL1
//		c = aux_adc[5];
////		c = (c >> 2) & 0x03FF;	// reduce to 10-bit
//		c >>= 2;	// reduce to 10-bit
//		DAC(c);
//	}
//}
//
//void LNA_AGC_poll(void){
//	// adjust the 1st attenuator after the LNA to prevent saturation,
//	// based on the RSSI measurement
//
//	// 2 AGC modes as set in REG39(0)
//	unsigned char agc_loop;
//	agc_loop = regv[39] & 0x01;
//
//
//	if (agc_loop == 0) {
//		// open loop. Gain is fixed.
//		return;
//	} else if (agc_loop == 1){
//		// LNA AGC closed loop algorithm relying on RSSI (prevent saturation at
//		// amplifiers), RSSI (RF_POWER_DET1) -> aux_adc[0]
//
//		// response: flat at 1023 from 0 to 0x800, then linear down to 0 at 0xFFF
//		// based on COM-0013 board measurements.
//		if(aux_adc[0] > 0x0800){
//			LNA_gain = 1023 - (((aux_adc[0] - 0x800) >> 1)& 0x3FF);
//		} else {
//			LNA_gain = 1023;
//		}
//		setLNAgain();	// set LNA gain
//	}
//}
//
//void RF_AGC_poll(void) {
//	// 2 AGC modes as set in REG39(1:0)
//	unsigned char agc_loop;
//	agc_loop = (regv[39] >> 1) & 0x01;
//
//	// check level/power sensors for reporting status (even when AGC loop is open) 3/9/13
//	if (agc_loop == 0) {
//		// open loop. Gain is fixed.
//		return;
//	} else if(((regv[39] & 0x0040) == 0) && ((FIO1PIN & 0x20000000) == 0)){
//		// freeze RF AGC when in half-duplex mode and RX_TXN = '0'
//		return;
//	} else if (agc_loop == 1){
//		// RF AGC closed loop algorithm relying on RF power detection (prevent saturation at
//		// RF mixer), RF_POWER_DET2 -> aux_adc[1]
//
//		// read RF power detect rf_power_det (done periodically in main loop)
//		// new: threshold reduced to prevent saturation (wideband noise + signal)
////		if ((aux_adc[1] > 0x0FA8) && (aux_adc[2] > 0x0E80)) {
//		if (aux_adc[1] > 0x0E74) {
//			// getting close to saturation at the RF mixer. Decrease gain
//			RF_gain_too_large_duration++;
//			RF_gain_too_small_duration = 0; // reset the time elapsed since last peak detect > 0x80
//			if (RF_gain_too_large_duration > 512) { // agc reaction time implied here
//				RF_gain_too_large_duration = 0; // reset counter
//				RF_gain_decrement();
//			}
//		} else {
//			// slow pressure to increase gain.
//			RF_gain_too_small_duration++; // increment the time elapsed since last saturation detect
//			if (RF_gain_too_small_duration > 512) { // agc reaction time implied here
//				// gain is consistently too small.
//				RF_gain_increment();
//				RF_gain_too_large_duration = 0; // reset counter
//				RF_gain_too_small_duration = 0; // reset the time elapsed since last saturation detect
//			}
//		}
//	}
//}
//
//void IF_AGC_poll(void) {
//	// 3 AGC modes as set in REG39(3:2)
//	unsigned char agc_loop;
//	agc_loop = (regv[39] & 0x0C) >> 2;
//
//	if (agc_loop == 0) {
//		// open loop. Gain is fixed.
//		return;
//	} else if(((regv[39] & 0x0040) == 0) && ((FIO1PIN & 0x20000000) == 0)){
//		// freeze IF AGC when in half-duplex mode and RX_TXN = '0'
//		return;
//	} else if (agc_loop == 1) {
//		// closed internal IF AGC loop based on local IF1_POWER_DET measurement
//		if (aux_adc[2] > 0x0E80) {
//			// Near saturation detected. Reduce gain
//			IF_gain_too_large_duration++;
//			IF_gain_too_small_duration = 0; // reset the time elapsed since last peak detect > 0x80
//			if (IF_gain_too_large_duration > 16) { // agc reaction time implied here
//				IF_gain_too_large_duration = 0; // reset counter
//				// decrement
//				if (IF_gain > 0) {
//					IF_gain--;
//					setIFGain();
//				} else {
//					IF_gain = 0; // enforce range 0 -  4095
//				}
//			}
//		} else {
//			// slow pressure to increase the gain
//			IF_gain_too_small_duration++; // increment the time elapsed since last saturation detect
//			if (IF_gain_too_small_duration > 16) { // agc reaction time implied here
//				// gain is consistently too small.
//				IF_gain_too_large_duration = 0; // reset counter
//				IF_gain_too_small_duration = 0; // reset the time elapsed since last saturation detect
//				// increment
//				if (IF_gain < 4095) {
//					IF_gain++;
//					setIFGain();
//				} else {
//					IF_gain = 4095; // enforce range 0 -  4095
//				}
//			}
//		}
//	} else if (agc_loop == 2) {
//		// external AGC loop. Controlled by AGC_IN (40-pin connector, A6)
//		// read ADC (converting the AGC_IN signal from the 40-pin connector to a 12-bit value)
//		// flip gain control value (ComBlocks AGC rule: 0 is for maximum gain)
//		IF_gain = 4095 - aux_adc[3];	// agc_in range 0 - 4095
//		// set digital gain
//		setIFGain();
//	}
//}

//void RF_gain_increment(void) {
//	if (RF_gain < 4095) {
//		RF_gain++;
//		setRFGain();
//	} else {
//		RF_gain = 4095; // enforce range 0 -  4095
//	}
//}
//
//void RF_gain_decrement(void) {
//	if (RF_gain > 0) {
//		RF_gain--;
//		setRFGain();
//	} else {
//		RF_gain = 0; // enforce range 0 -  4095
//	}
//}

//long pinDiodeAttenuatorCompensation(long gain){
//	// attempt at avoiding instabilities
//	// function of center frequency fc (global)
//	// input range 0 - 1023
//
//	long revised_gain;
//	revised_gain = gain;
//
//	if(fc < 200000000){
//		// below 200 MHz, avoid very large attenuations (Vc < 0.55V) which are destabilizing
//		if(gain < 129)
//			revised_gain = 129;
//	} else if (fc < 500000000){
//		// below 500 MHz, avoid very large attenuations (Vc < 0.5V) which are destabilizing
//		if(gain < 108)
//			revised_gain = 108;
//	} else if (fc > 2000000000){
//		// above 2 GHz, avoid too much gain (> 2.5V) because it can create instabilities
//		if(gain > 538)
//			revised_gain = 538;
//	}
//	return revised_gain;
//}

void ADC_init(void) {
	// ADC initialization
	PCONP |= 0x00001000; // power ADC subsystem
	// PCLKSEL0: by default peripheral clock is CCLK/4 = 30 MHz
	// A/D Control Register: inputs are AD0[2] and [5], ADC clock is 10 MHz, burst=1, operational,
	AD0CR = 0x00210224;
}

void ADC_TX_POWER_DET(void) {
	// returns the last value read on TX_POWER_DET = AD0[5]
	unsigned long a;
	a = AD0DR5;
	if (a & 0x80000000) {
		// done bit high. update tx_power_det
		tx_power_det = (a >> 4) & 0x00000FFF; // 12-bit ADC measurement
	}
}

void ADC_TEMP(void) {
	// returns the last value read on TEMP = AD0[2]
	unsigned long a;
	a = AD0DR2;
	if (a & 0x80000000) {
		// done bit high. update temperature_det
		temperature_det = (a >> 4) & 0x00000FFF; // 12-bit ADC measurement
	}
}

void TP_poll(void){
	// forward selected status register to analog test point
	if((regv[6] & 0x03) == 1){
		// TP = temperature sensor
		DAC(temperature_det >> 2);
	} else if((regv[6] & 0x03) == 2){
		// TP = transmit power sensor
		DAC(tx_power_det >> 2);
	}

}

//void setLPF() {
//	// set the low-pass filter bandwidth
//	// REG4/REG5 stores the one-sided bandwidth in kHz
//	// Controls the CIC decimation filter and the two Half-Band filters
//	int bw, cic_r;
//
//	// read single-sided bandwidth in KHz
//	bw = regv[5] & 0x00FF;
//	bw = ((bw << 8) & 0x00FF00) | (regv[4] & 0x00FF);
//
//	// assuming a 100 MSamples/s sampling clock at the FPGA, let's calculate the
//	// decimation and the need to bypass
//	// we want to decimate from +/-fs/2 (at the ADC) to +/-4*bw (at the CIC filter output)
//	cic_r = (int) (50000 / (4 * bw));
//	// check valid range 1 - 16384
//	if (cic_r < 1)
//		cic_r = 1; // 0 is not an acceptable value
//	else if (cic_r > 16384)
//		cic_r = 16384;
//	regv[107] = (unsigned char) (cic_r & 0x00FF);
//	regv[108] = (unsigned char) ((cic_r >> 8) & 0x003F);
//
//	// bypass filters
//	if (bw > 12500) {
//		// bypass CIC + 1st HBF
//		regv[108] = (regv[108] & 0xC0) | 0x40;
//	} else if (bw > 6250) {
//		// bypass CIC
//		regv[108] = (regv[108] & 0xC0) | 0x80;
//	} else {
//		// enable all filters
//		regv[108] = (regv[108] & 0xC0) | 0xC0;
//	}
//
//	fpga_setByte(107, regv[107]);
//	fpga_setByte(108, regv[108]);
//
//}

//int AD7298_rw(int cw ){
//	// write control word cw, read data from ADC
//	// write control word = 0 when no control word to write
//	// returns 4 bit address + 12-bit sample
//	// usage: write FCD2 once to enable repetitive measurements on all channels except 7
//	int i, retval;
//	retval = 0;
//
//	PIN_AD7298_SCLK_H();
//	PIN_AD7298_CSB_L();
//	for(i = 0; i < 16; i++){
//		// read bit from DOUT
//		retval <<= 1;	// MSb first when reading
//		if(FIO2PIN & 0x080){	// P2[7]
//			retval |= 0x00001;
//		} else {
//			retval &= ~0x00001;
//		}
//		// write bit to DIN
//		if(cw & 0x08000){
//			PIN_AD7298_DIN_H();
//		} else {
//			PIN_AD7298_DIN_L();
//		}
//		cw <<= 1;	// MSb first when writing
//		PIN_AD7298_SCLK_L();
//		PIN_AD7298_SCLK_H();
//
//	}
//	PIN_AD7298_CSB_H();
//	return retval;
//}


//void AD5621_1(int dac_val){
//	// 12-bit auxiliary DAC (faster than internal DAC and PWMs)
//	int i;
//	dac_val = (dac_val << 2) & 0x3FFC;	// prepend 2 'normal operation' bits. append 2 x bits.
//
//	PIN_AD5621_1_SYNCB_H();
//	PIN_AD5621_1_SCLK_H();
//	PIN_AD5621_1_SDIN_L();
//
//	PIN_AD5621_1_SYNCB_L();
//	for (i = 0; i < 16; i++) {
//		if(dac_val & 0x8000){
//			PIN_AD5621_1_SDIN_H();
//		} else {
//			PIN_AD5621_1_SDIN_L();
//		}
//		PIN_AD5621_1_SCLK_L();
//		dac_val <<= 1;
//		PIN_AD5621_1_SCLK_H();
//	}
//	PIN_AD5621_1_SYNCB_H();
//}

//void AD5621_2(int dac_val){
//	// 12-bit auxiliary DAC (faster than internal DAC and PWMs)
//	int i;
//	dac_val = (dac_val << 2) & 0x3FFC;	// prepend 2 'normal operation' bits. append 2 x bits.
//
//	PIN_AD5621_2_SYNCB_H();
//	PIN_AD5621_2_SCLK_H();
//	PIN_AD5621_2_SDIN_L();
//
//	PIN_AD5621_2_SYNCB_L();
//	for (i = 0; i < 16; i++) {
//		if(dac_val & 0x8000){
//			PIN_AD5621_2_SDIN_H();
//		} else {
//			PIN_AD5621_2_SDIN_L();
//		}
//		PIN_AD5621_2_SCLK_L();
//		dac_val <<= 1;
//		PIN_AD5621_2_SCLK_H();
//	}
//	PIN_AD5621_2_SYNCB_H();
//}

